"""Grid-like Contrail Cirrus Prediction (CoCiP) modeling support."""

from pycontrails.models.cocipgrid.cocip_grid import CocipGrid
from pycontrails.models.cocipgrid.cocip_grid_params import CocipGridParams

__all__ = [
    "CocipGrid",
    "CocipGridParams",
]
